// Copyright (C) 2017 Luke Shumaker <lukeshu@sbcglobal.net>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301 USA

package nslcd_server

import (
	"context"

	"git.lukeshu.com/go/libsystemd/sd_daemon"
	"golang.org/x/sys/unix"
)

// Logger is the common interface between
// `"git.lukeshu.com/go/libsystemd/sd_daemon".Logger` and
// `"log/syslog".Writer`.
type Logger interface {
	Emerg(m string) error
	Alert(m string) error
	Crit(m string) error
	Err(m string) error
	Warning(m string) error
	Notice(m string) error
	Info(m string) error
	Debug(m string) error

	Write(m []byte) (int, error)
}

// contextKey is a value for use with context.WithValue.  It's used as
// a pointer so it fits in an interface{} without allocation.
type contextKey struct {
	name string
}

var (
	// PeerCredKey is a context key.  It can be used in backend
	// methods to access the credentials of the client process.
	// The associated value will be of type
	// "golang.org/x/sys/unix".Ucred
	PeerCredKey = &contextKey{"peercred"}

	// LoggerKey is a context key.  It can be used in the backend
	// methods to access a logger.  The associated value will be
	// an implementation of Logger.
	LoggerKey = &contextKey{"log"}
)

// PeerCredFromContext is a convenience function for
//
//     cred, ok := ctx.Value(nslcd_server.PeerCredKey).(unix.Ucred)
func PeerCredFromContext(ctx context.Context) (unix.Ucred, bool) {
	cred, ok := ctx.Value(PeerCredKey).(unix.Ucred)
	return cred, ok
}

func LoggerFromContext(ctx context.Context) Logger {
	logger, ok := ctx.Value(LoggerKey).(Logger)
	if !ok {
		return sd_daemon.Log
	}
	return logger
}
